@@ -9,11 +9,12 @@
 # Get this list via:
 #    using BinaryBuilder
 #    print("TRIPLETS=\"$(join(triplet.(BinaryBuilder.supported_platforms()), " "))\"")
-TRIPLETS="i686-linux-gnu x86_64-linux-gnu aarch64-linux-gnu arm-linux-gnueabihf powerpc64le-linux-gnu i686-linux-musl x86_64-linux-musl aarch64-linux-musl arm-linux-musleabihf x86_64-apple-darwin14 x86_64-unknown-freebsd11.1 i686-w64-mingw32 x86_64-w64-mingw32"
+TRIPLETS="i686-linux-gnu x86_64-linux-gnu aarch64-linux-gnu armv7l-linux-gnueabihf powerpc64le-linux-gnu i686-linux-musl x86_64-linux-musl aarch64-linux-musl armv7l-linux-musleabihf x86_64-apple-darwin14 x86_64-unknown-freebsd11.1 i686-w64-mingw32 x86_64-w64-mingw32"
 
 # These are the projects currently using BinaryBuilder; both GCC-expanded and non-GCC-expanded:
-BB_PROJECTS="gmp mbedtls libssh2 mpfr curl libgit2 pcre libuv unwind osxunwind dsfmt objconv p7zip zlib"
-BB_GCC_EXPANDED_PROJECTS="llvm openblas suitesparse openlibm"
+BB_PROJECTS="gmp mbedtls libssh2 mpfr curl libgit2 pcre libuv unwind osxunwind dsfmt objconv p7zip zlib suitesparse openlibm"
+BB_GCC_EXPANDED_PROJECTS="openblas"
+BB_CXX_EXPANDED_PROJECTS="llvm"
 
 # If we've been given a project name, filter down to that one:
 if [ -n "${1}" ]; then
@@ -25,6 +26,10 @@ if [ -n "${1}" ]; then
         *${1}*) BB_GCC_EXPANDED_PROJECTS="${1}" ;;
         *) BB_GCC_EXPANDED_PROJECTS="" ;;
     esac
+    case "${BB_CXX_EXPANDED_PROJECTS}" in
+        *${1}*) BB_CXX_EXPANDED_PROJECTS="${1}" ;;
+        *) BB_CXX_EXPANDED_PROJECTS="" ;;
+    esac
 fi
 
 # Get "contrib/" directory path
@@ -40,9 +45,17 @@ for triplet in ${TRIPLETS}; do
 
     for proj in ${BB_GCC_EXPANDED_PROJECTS}; do
 		PROJ="$(echo ${proj} | tr [a-z] [A-Z])"
-        for gcc in gcc4 gcc7 gcc8; do
-		    make -C "${CONTRIB_DIR}/../deps" USE_BINARYBUILDER_${PROJ}=1 ${PROJ}_BB_TRIPLET=${triplet}-${gcc} BB_TRIPLET_CXXABI=${triplet} distclean-${proj}
-		    make -C "${CONTRIB_DIR}/../deps" USE_BINARYBUILDER_${PROJ}=1 ${PROJ}_BB_TRIPLET=${triplet}-${gcc} BB_TRIPLET_CXXABI=${triplet} install-${proj}
+        for libgfortran in libgfortran3 libgfortran4 libgfortran5; do
+		    make -C "${CONTRIB_DIR}/../deps" USE_BINARYBUILDER_${PROJ}=1 ${PROJ}_BB_TRIPLET=${triplet}-${libgfortran} BB_TRIPLET_CXXABI=${triplet} distclean-${proj}
+		    make -C "${CONTRIB_DIR}/../deps" USE_BINARYBUILDER_${PROJ}=1 ${PROJ}_BB_TRIPLET=${triplet}-${libgfortran} BB_TRIPLET_CXXABI=${triplet} install-${proj}
+        done
+    done
+
+    for proj in ${BB_CXX_EXPANDED_PROJECTS}; do
+		PROJ="$(echo ${proj} | tr [a-z] [A-Z])"
+        for cxx in cxx03 cxx11; do
+		    make -C "${CONTRIB_DIR}/../deps" USE_BINARYBUILDER_${PROJ}=1 ${PROJ}_BB_TRIPLET=${triplet}-${cxx} BB_TRIPLET_CXXABI=${triplet} distclean-${proj}
+		    make -C "${CONTRIB_DIR}/../deps" USE_BINARYBUILDER_${PROJ}=1 ${PROJ}_BB_TRIPLET=${triplet}-${cxx} BB_TRIPLET_CXXABI=${triplet} install-${proj}
         done
     done
 done
