@@ -29,6 +29,7 @@
 #ifndef __UNWINDCURSOR_HPP__
 #define __UNWINDCURSOR_HPP__
 
+#include <assert.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -405,7 +406,8 @@ template <typename A, typename R>
 UnwindCursor<A,R>::UnwindCursor(unw_context_t* context, A& as)
   : fUnwindInfoMissing(false), fIsSignalFrame(false), fForceDwarf(false), fRegisters(context), fAddressSpace(as)
 {
-	COMPILE_TIME_ASSERT( sizeof(UnwindCursor<A,R>) < sizeof(unw_cursor_t) );
+	static_assert( sizeof(UnwindCursor<A,R>) < sizeof(unw_cursor_t),
+	"sizeof(UnwindCursor<A,R>) MUST be < sizeof(unw_cursor_t)" );
 
 	bzero(&fInfo, sizeof(fInfo));
 }
